"""
Contains all functions used for searching Stack Overflow and Google
"""

import os
import random
import re
import sys
import copy

from bs4 import BeautifulSoup
import requests
import urwid

import socli.printer
import socli.tui

uas = []  # User agent list
header = {}  # Request header
google_search = True
so_url = "http://stackoverflow.com"  # Site URL
so_qurl = "http://stackoverflow.com/search?q="  # Query URL
so_burl = "https://stackoverflow.com/?tab="  # Assuming browse URL
google_search_url = "https://www.google.com/search?q=site:stackoverflow.com+"  # Google search query URL


def get_questions_for_query(query, count=10):
    """
    Fetch questions for a query using Stack Overflow default search mechanism.
    Returned question urls are relative to SO homepage.

    :param count: At most 10 questions are returned. (Can be altered by passing count)
    :param query: User-entered query string
    :return: list of [ (question_text, question_description, question_url) ]
    """

    questions = []
    random_headers()

    search_res = requests.get(so_qurl + query, headers=header)
    captcha_check(search_res.url)
    soup = BeautifulSoup(search_res.text, 'html.parser')
    try:
        soup.find_all("div", class_="question-summary")[0]  # For explicitly raising exception
    except IndexError:
        socli.printer.print_warning("No results found...")
        sys.exit(0)
    tmp = (soup.find_all("div", class_="question-summary"))
    tmp1 = (soup.find_all("div", class_="excerpt"))
    i = 0
    while i < len(tmp) and i < count:  # iterate over and limit results
        question_text = ' '.join((tmp[i].a.get_text()).split())
        question_text = question_text.replace("Q: ", "")
        question_desc = (tmp1[i].get_text()).replace("'\r\n", "")
        question_desc = ' '.join(question_desc.split())
        question_local_url = tmp[i].a.get("href")
        questions.append((question_text, question_desc, question_local_url))
        i += 1

    return questions


def get_questions_for_query_google(query, count=10):
    """
    Fetch questions for a query using Google search.
    Returned question urls are URLS to SO homepage.
    At most 10 questions are returned. (Can be altered by passing count)
    :param query: User-entered query string
    :return: list of [ (question_text, question_description, question_url) ]
    """
    i = 0
    questions = []
    random_headers()
    search_results = requests.get(google_search_url + query, headers=header)
    captcha_check(search_results.url)
    soup = BeautifulSoup(search_results.text, 'html.parser')
    try:
        soup.find_all("div", class_="g")[0]  # For explicitly raising exception
    except IndexError:
        socli.printer.print_warning("No results found...")
        sys.exit(0)
    for result in soup.find_all("div", class_="g"):
        if i == count:
            break
        try:
            question_title = result.find("h3").get_text().replace(' - Stack Overflow', '')
            question_desc = result.find("div", recursive=False).find_all("div", recursive=False)[1] \
                .find_all("span")[-2].get_text()
            if question_desc == "":  # For avoiding instant answers
                raise NameError  # Explicit raising
            question_url = result.find("a").get("href")  # Retrieves the Stack Overflow link
            question_url = fix_google_url(question_url.lower())

            if question_url is None:
                i = i - 1
                continue

            questions.append([question_title, question_desc, question_url])
            i += 1
        except NameError:
            continue
        except AttributeError:
            continue

    # Check if there are any valid question posts
    if not questions:
        socli.printer.print_warning("No results found...")
        sys.exit(0)
    return questions


def get_comments(soup):
    """
    :param soup
    extacts comments on the answers
    """
    comments_list = []
    raw_comments_list = soup.find_all("ul", class_="js-comments-list")
    for raw_comments in raw_comments_list:
        comments_list.append([str(index + 1) + ' ' + raw_comment.get_text() + '\n' for index, raw_comment in
                              enumerate(raw_comments.find_all("span", class_='comment-copy'))])
    if len(comments_list) <= 1:
        comments_list.append(["No comments as there are no answers to this question..."])
    return comments_list[1::]


def get_question_stats_and_answer_and_comments(url):
    """
    Fetch the content of a StackOverflow page for a particular question.
    :param url: full url of a StackOverflow question
    :return: tuple of ( question_title, question_desc, question_stats, answers, comments, dup_url )
    """
    random_headers()
    res_page = requests.get(url, headers=header)
    captcha_check(res_page.url)
    soup = BeautifulSoup(res_page.text, 'html.parser')
    dup_url = None
    question_title, question_desc, question_stats, dup_url = get_stats(soup)
    answers = [s.get_text() for s in soup.find_all("div", class_="js-post-body")][
              1:]  # first post is question, discard it.
    accepted_answer  = soup.find_all("div",class_="accepted-answer")[0].find_all("div",class_="js-post-body")[0].get_text()
    if accepted_answer in answers:
        answers.remove(accepted_answer)
    accepted_answer = "=============ACCEPTED_ANSWER===========\n" + accepted_answer + "\n===============ACCEPTED_ANSWER============"
    answers.insert(0,accepted_answer)
    comments = get_comments(soup)
    if len(answers) == 0:
        answers.append('No answers for this question ...')
    return question_title, question_desc, question_stats, answers, comments, dup_url


def get_stats(soup):
    """
    Get Question stats
    :param soup:
    :return: tuple of (question_title, question_desc, question_stats, dup_url)
    """
    question_title = (soup.find_all("a", class_="question-hyperlink")[0].get_text())
    question_stats = (soup.find_all("div", class_="js-vote-count")[0].get_text())
    dup_url = None
    try:

        asked_info = soup.find("time").parent.get_text()
        active_info = soup.find("time").parent.findNext('div').get_text()
        viewed_info = soup.find("time").parent.findNext('div').findNext('div').get_text()
        question_stats = "Votes " + question_stats + " | " + asked_info + " | " + active_info + " | " + viewed_info
    except:
        question_stats = "Could not load statistics."
    question_desc = (soup.find_all("div", class_="js-post-body")[0])
    if '[duplicate]' in question_title:
        dup_answer = (soup.find_all("div", class_="js-post-body")[0])
        link = dup_answer.find('a')['href']
        link = so_url + link
        dup_url = copy.deepcopy(link)
        # using deepcopy else after the decompose of the first div, the url will be lost.
        question_desc.find('div').decompose()
    add_urls(question_desc)
    question_desc = question_desc.get_text()
    question_stats = ' '.join(question_stats.split())
    return question_title, question_desc, question_stats, dup_url


def add_urls(tags):
    """
    Adds the URL to any hyperlinked text found in a question
    or answer.
    :param tags:
    """
    images = tags.find_all("a")

    for image in images:
        if hasattr(image, "href"):
            image.string = "{} [{}]".format(image.text, image['href'])


def socli_interactive_windows(query):
    """
    Interactive mode basic implementation for windows, since urwid doesn't support CMD.
    :param query:
    :return:
    """
    try:
        search_res = requests.get(so_qurl + query)
        captcha_check(search_res.url)
        soup = BeautifulSoup(search_res.text, 'html.parser')
        try:
            soup.find_all("div", class_="question-summary")[0]  # For explictly raising exception
            tmp = (soup.find_all("div", class_="question-summary"))
            tmp1 = (soup.find_all("div", class_="excerpt"))
            i = 0
            question_local_url = []
            print(socli.printer.bold("\nSelect a question below:\n"))
            while i < len(tmp):
                if i == 10:
                    break  # limiting results
                question_text = ' '.join((tmp[i].a.get_text()).split())
                question_text = question_text.replace("Q: ", "")
                question_desc = (tmp1[i].get_text()).replace("'\r\n", "")
                question_desc = ' '.join(question_desc.split())
                socli.printer.print_warning(str(i + 1) + ". " + socli.printer.display_str(question_text))
                question_local_url.append(tmp[i].a.get("href"))
                print("  " + socli.printer.display_str(question_desc) + "\n")
                i = i + 1
            try:
                op = int(socli.printer.inputs("\nType the option number to continue or any other key to exit:"))
                while 1:
                    if (op > 0) and (op <= i):
                        socli.printer.display_results(so_url + question_local_url[op - 1])
                        cnt = 1  # this is because the 1st post is the question itself
                        while 1:
                            global tmpsoup
                            qna = socli.printer.inputs(
                                "Type " + socli.printer.bold("o") + " to open in browser, " + socli.printer.bold(
                                    "n") + " to next answer, " + socli.printer.bold(
                                    "b") + " for previous answer or any other key to exit:")
                            if qna in ["n", "N"]:
                                try:
                                    answer = (tmpsoup.find_all("div", class_="js-post-body")[cnt + 1].get_text())
                                    socli.printer.print_green("\n\nAnswer:\n")
                                    print("-------\n" + answer + "\n-------\n")
                                    cnt = cnt + 1
                                except IndexError:
                                    socli.printer.print_warning(" No more answers found for this question. Exiting...")
                                    sys.exit(0)
                                continue
                            if qna in ["b", "B"]:
                                if cnt == 1:
                                    socli.printer.print_warning(
                                        " You cant go further back. You are on the first answer!")
                                    continue
                                answer = (tmpsoup.find_all("div", class_="js-post-body")[cnt - 1].get_text())
                                socli.printer.print_green("\n\nAnswer:\n")
                                print("-------\n" + answer + "\n-------\n")
                                cnt = cnt - 1
                                continue
                            if qna in ["o", "O"]:
                                import webbrowser
                                if sys.platform.startswith('darwin'):
                                    browser = webbrowser.get('safari')
                                else:
                                    browser = webbrowser.get()
                                socli.printer.print_warning("Opening in your browser...")
                                browser.open(so_url + question_local_url[op - 1])
                            else:
                                break
                        sys.exit(0)
                    else:
                        op = int(input("\n\nWrong option. select the option no to continue:"))
            except Exception as e:
                socli.printer.showerror(e)
                socli.printer.print_warning("\n Exiting...")
                sys.exit(0)
        except IndexError:
            socli.printer.print_warning("No results found...")
            sys.exit(0)

    except UnicodeEncodeError:
        socli.printer.print_warning("\n\nEncoding error: Use \"chcp 65001\" command before using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        socli.printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        socli.printer.showerror(e)
        sys.exit(0)


def socli_interactive(query):
    """
    Interactive mode
    :return:
    """
    if sys.platform == 'win32':
        return socli_interactive_windows(query)

    class SelectQuestionPage(urwid.WidgetWrap):

        def display_text(self, index, question):
            question_text, question_desc, _ = question
            text = [
                ("warning", u"{}. {}\n".format(index, question_text)),
                question_desc + "\n",
            ]
            return text

        def __init__(self, questions):
            self.questions = questions
            self.cachedQuestions = [None for _ in range(10)]
            widgets = [self.display_text(i, q) for i, q in enumerate(questions)]
            self.questions_box = socli.tui.ScrollableTextBox(widgets)
            self.header = socli.tui.UnicodeText(('less-important', 'Select a question below:\n'))
            self.footerText = '0-' + str(len(self.questions) - 1) + ': select a question, any other key: exit.'
            self.errorText = socli.tui.UnicodeText.to_unicode('Question numbers range from 0-' +
                                                              str(len(self.questions) - 1) +
                                                              ". Please select a valid question number.")
            self.footer = socli.tui.UnicodeText(self.footerText)
            self.footerText = socli.tui.UnicodeText.to_unicode(self.footerText)
            frame = urwid.Frame(header=self.header,
                                body=urwid.Filler(self.questions_box, height=('relative', 100), valign='top'),
                                footer=self.footer)
            urwid.WidgetWrap.__init__(self, frame)

        # Override parent method
        def selectable(self):
            return True

        def keypress(self, size, key):
            if key in '0123456789':
                try:
                    question_url = self.questions[int(key)][2]
                    self.footer.set_text(self.footerText)
                    self.select_question(question_url, int(key))
                except IndexError:
                    self.footer.set_text(self.errorText)
            elif key in {'down', 'up'}:
                self.questions_box.keypress(size, key)
            else:
                raise urwid.ExitMainLoop()

        def select_question(self, url, index):
            if self.cachedQuestions[index] is not None:
                socli.tui.question_post = self.cachedQuestions[index]
                socli.tui.MAIN_LOOP.widget = socli.tui.question_post
            else:
                if not google_search:
                    url = so_url + url
                question_title, question_desc, question_stats, answers, comments, dup_url = get_question_stats_and_answer_and_comments(url)
                socli.tui.question_post = socli.tui.QuestionPage(
                    (url, question_title, question_desc, question_stats, answers, comments, dup_url, None))
                self.cachedQuestions[index] = socli.tui.question_post
                socli.tui.MAIN_LOOP.widget = socli.tui.question_post

    socli.tui.display_header = socli.tui.Header()

    try:
        if google_search:
            questions = get_questions_for_query_google(query)
        else:
            questions = get_questions_for_query(query)
        question_page = SelectQuestionPage(questions)
        socli.tui.MAIN_LOOP = socli.tui.EditedMainLoop(question_page, socli.printer.palette)
        socli.tui.MAIN_LOOP.run()

    except UnicodeEncodeError:
        socli.printer.print_warning("\n\nEncoding error: Use \"chcp 65001\" command before using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        socli.printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        socli.printer.showerror(e)
        print("exiting...")
        sys.exit(0)


def socli_manual_search(query, rn):
    """
    Manual search by question index
    :param query:
    :param rn:
    :return:
    """
    if rn < 1:
        socli.printer.print_warning(
            "Count starts from 1. Use: \"socli -i 2 -q python for loop\" for the 2nd result for the query")
        sys.exit(0)
    query = socli.printer.urlencode(query)
    try:
        random_headers()
        # Set count = 99 so you can choose question numbers higher than 10
        count = 99
        res_url = None
        try:
            if google_search:
                questions = get_questions_for_query_google(query, count)
                res_url = questions[rn - 1][2]
            else:
                questions = get_questions_for_query(query, count)
                res_url = so_url + questions[rn - 1][2]
            socli.printer.display_results(res_url)
        except IndexError:
            socli.printer.print_warning("No results found...")
            sys.exit(1)
    except UnicodeEncodeError:
        socli.printer.print_warning("Encoding error: Use \"chcp 65001\" command before "
                                    "using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        socli.printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        socli.printer.showerror(e)
        sys.exit(0)


def fix_google_url(url):
    """
    Fixes the url extracted from HTML when
    performing a google search
    :param url:
    :return: Correctly formatted URL to be used in requests.get
    """
    if "&sa=" in url:
        url = url.split("&")[0]
    if "/url?q=" in url[0:7]:
        url = url[7:]  # Removes the "/url?q=" prefix

    if url[:30] == "http://www.google.com/url?url=":
        # Used to get rid of this header and just retrieve the Stack Overflow link
        url = url[30:]

    if "http" not in url[:4]:
        url = "https://" + url  # Add the protocol if it doesn't already exist

    # Makes sure that we stay in the questions section of Stack Overflow
    if not bool(re.search(r"/questions/[0-9]+", url)) and not bool(re.search(r"\.com/a/[0-9]", url)):
        return None

    if url[:17] == "https:///url?url=":  # Resolves rare bug in which this is a prefix
        url = url[17:]

    return url


def captcha_check(url):
    """
    Exits program when there is a captcha, preventing any error.
    Users will have to manually verify their identity.
    :param url: URL from Stack Overflow
    :return:
    """
    if google_search:
        google_error_display_msg = "Google thinks you're a bot because you're issuing too many queries too quickly!" + \
                                   " Now you'll have to wait about an hour before you're unblocked... :(. " \
                                   "Use the -s tag to search via Stack Overflow instead."
        # Check if google detects user as a bot
        if re.search(r"ipv4\.google\.com/sorry", url):
            socli.printer.print_warning(google_error_display_msg)
            exit(0)
    else:
        if re.search(r"\.com/nocaptcha", url):  # Searching for stackoverflow captcha check
            socli.printer.print_warning(
                "StackOverflow captcha check triggered. Please wait a few seconds before trying again. "
                "Try solving it manually: " + url)
            exit(0)


def load_user_agents():
    """
    Loads the list of user agents from user_agents.txt
    :return:
    """
    global uas
    uas = []
    with open(os.path.join(os.path.dirname(__file__), "user_agents.txt"), 'rb') as uaf:
        for ua in uaf.readlines():
            if ua:
                uas.append(ua.rstrip())
    random.shuffle(uas)


def random_headers():
    """
    Sets header variable to a random value
    :return:
    """
    global uas
    global header
    if not uas:
        load_user_agents()
    ua = random.choice(uas)
    header = {"User-Agent": ua}



# Initialize Sentry
import sentry_sdk
from socli.version import __version__

sentry_sdk.init(
    "https://95c4106659044cbda2ea0fe499f4be7e@o323465.ingest.sentry.io/5445901",
    traces_sample_rate=1.0,
    release="socli-" + __version__
)

from socli.socli import main
main()



"""
Contains all functions used for parsing the command line args
"""
import argparse,argcomplete
import textwrap


def parse_arguments(command):
    """
    Parses the command into arguments and flags
    :param command: the command in list form
    :return: an object that contains the values for all arguments

    Currently, all help messages are not in use and helpman() is the default.
    Need to implement nicer --help format in the future.
    """
    parser = argparse.ArgumentParser(description=textwrap.dedent('Stack Overflow command line client'), add_help=False)

    # Comment this line out if you want to use argparse's default help function
    parser.add_argument('--help', '-h', action='store_true', help='Show this help message and exit')

    # Flags that return true if present and false if not
    parser.add_argument('--new', '-n', action='store_true',
                        help=textwrap.dedent("Opens the stack overflow new questions page in your "
                                             "default browser. You can create a new question using it."))
    parser.add_argument('--interactive', '-i', action='store_true', help=textwrap.dedent(
        "To search in Stack Overflow and display the matching results. "
        "You can choose and browse any of the results interactively"))
    parser.add_argument('--debug', action='store_true', help="Turn debugging mode on")
    parser.add_argument('--sosearch', '-s', action='store_true',
                        help="Searches directly on Stack Overflow instead of using Google")
    parser.add_argument('--api', '-a', action='store_true', help="Sets a custom API key for socli")
    parser.add_argument('--delete', '-d', action='store_true',
                        help="Deletes the configuration file generated by socli -u command")

    # Accepts 1 argument. Returns None if flag is not present and
    # 'STORED_USER' if flag is present, but no argument is supplied
    parser.add_argument('--user', '-u', nargs='?', const='(RANDOM_STRING_CONSTANT)', type=str,
                        help="Displays information about the user "
                             "provided as the next argument(optional). If no argument is provided "
                             "it will ask the user to enter a default username. Now the user "
                             "can run the command without the argument")

    # Accepts one or more arguments
    parser.add_argument('--tag', '-t', nargs='+', help="To search a query by tag on stack overflow."
                                                       "Visit http://stackoverflow.com/tags to see the list of all "
                                                       "tags.\n   eg:- socli --tag javascript,node.js --query "
                                                       "foo bar: Displays the search result of the query"
                                                       " \"foo bar\" in stack overflow's javascript and node.js tags")
    parser.add_argument('--query', '-q', nargs='+', default=[],
                        help="If any of the following commands are used then you "
                             "must specify this option and a query following it.")
    parser.add_argument('--browse', '-b', nargs='+', default=[],
                        help="Searches for ten hot,week and interesting questions on today's page.")
    # Accepts 0 or more arguments. Used to catch query if no flags are present
    parser.add_argument('userQuery', nargs='*', help=argparse.SUPPRESS)

    # Accepts 1 argument
    parser.add_argument('--res', '-r', type=int, help="To select and display a result manually and display "
                                                      "its most voted answer. \n   eg:- socli --res 2 --query "
                                                      "foo bar: Displays the second search result of the query"
                                                      " \"foo bar\"'s most voted answer")
    parser.add_argument('--open-url', '-o', nargs=1, type=str, help='To load the given url')
    parser.add_argument('--json', '-j', action='store_true', help='Write output to stdout as json')
    parser.add_argument('--register', '-g', action='store_true', help='Registers socli\'s shell autocompletion')
    parser.add_argument('--version' , '-v', action='store_true', help='Prints the current version of socli')
    
    argcomplete.autocomplete(parser)
    namespace = parser.parse_args(command)
    return namespace



"""
Urwid-based class hierarchy that forms the front end of the SoCLI application.
"""

import sys
import subprocess
import urwid

import socli.printer as pr

question_post = None
question_page = None
display_header = None
MAIN_LOOP = None


class UnicodeText(urwid.Text):
    """ encode all text to utf-8 """

    def __init__(self, text):
        # As we were encoding all text to utf-8 in output before with dispstr, do it automatically for all input
        text = UnicodeText.to_unicode(text)
        urwid.Text.__init__(self, text)

    @classmethod
    def to_unicode(cls, markup):
        """convert urwid text markup object to utf-8"""
        try:
            return pr.display_str(markup)
        except AttributeError:
            mapped = [cls.to_unicode(i) for i in markup]
            if isinstance(markup, tuple):
                return tuple(mapped)
            return mapped


class Header(UnicodeText):
    """
    Header of the question page. Event messages are recorded here.
    """

    def __init__(self):
        self.current_event = None
        UnicodeText.__init__(self, '')

    def event(self, event, message):
        self.current_event = event
        self.set_text(message)

    def clear(self, event):
        if self.current_event == event:
            self.set_text('')


class EditedMainLoop(urwid.MainLoop):

    def process_input(self, keys):
        super(EditedMainLoop, self).process_input(keys)
        global question_post
        if question_post is not None:
            if 'window resize' in keys:
                question_post.keypress(question_post, 'window resize')


class QuestionPage(urwid.WidgetWrap):
    """
    Main container for urwid interactive mode.
    """

    def __init__(self, data):
        """
        Construct the Question Page.
        :param data: tuple of (question_url, question_title, question_desc, question_stats, answers, comments, dup_url, dup_link)
        """
        question_url, question_title, question_desc, question_stats, answers, comments, dup_url, dup_link = data
        self.dup_url = dup_url
        self.dup_link = dup_link
        self.question_title = question_title
        self.question_desc = question_desc
        self.question_stats = question_stats
        self.url = question_url
        self.answer_text = AnswerText(answers, comments)
        answer_frame = self.make_frame()
        urwid.WidgetWrap.__init__(self, answer_frame)

    def make_frame(self):
        """
        Returns a new frame that is formatted correctly with respect to the window's dimensions.
        :return: a new urwid.Frame object
        """
        self.screenHeight, screenWidth = subprocess.check_output(
            ['stty', 'size']).split()
        self.question_text = urwid.BoxAdapter(QuestionDescription(self.question_desc),
                                              int(max(1, (int(self.screenHeight) - 9) / 2)))
        if self.dup_url:
            answer_frame = urwid.Frame(
                header=urwid.Pile([
                    display_header,
                    QuestionTitle(self.question_title),
                    self.question_text,
                    QuestionStats(self.question_stats),
                    urwid.Divider('-')
                ]),
                body=self.answer_text,
                footer=urwid.Pile([
                    QuestionURL(self.url),
                    UnicodeText(
                        u'\u2191: previous answer, \u2193: next answer, c:comments, o: open in browser, \u2190: back, '
                        u'd: visit duplicated question, q: quit')
                ])
            )
        elif self.dup_link:
            answer_frame = urwid.Frame(
                header=urwid.Pile([
                    display_header,
                    QuestionTitle(self.question_title),
                    self.question_text,
                    QuestionStats(self.question_stats),
                    urwid.Divider('-')
                ]),
                body=self.answer_text,
                footer=urwid.Pile([
                    QuestionURL(self.url),
                    UnicodeText(
                        u'\u2191: previous answer, \u2193: next answer, c:comments, o: open in browser, \u2190: back, '
                        u'd: back to original question, q: quit')
                ])
            )
        else:
            answer_frame = urwid.Frame(
                header=urwid.Pile([
                    display_header,
                    QuestionTitle(self.question_title),
                    self.question_text,
                    QuestionStats(self.question_stats),
                    urwid.Divider('-')
                ]),
                body=self.answer_text,
                footer=urwid.Pile([
                    QuestionURL(self.url),
                    UnicodeText(
                        u'\u2191: previous answer, \u2193: next answer, c: comments, o: open in browser, '
                        u'\u2190: back, q: quit')
                ])
            )
        return answer_frame

    def make_comment_frame(self):
        """
        Returns a new frame that is formatted correctly with respect to the window's dimensions.
        :return: a new urwid.Frame object
        """
        self.screenHeight, screenWidth = subprocess.check_output(
            ['stty', 'size']).split()
        self.question_text = urwid.BoxAdapter(QuestionDescription(self.question_desc),
                                              int(max(1, (int(self.screenHeight) - 9) / 2)))

        comment_frame = urwid.Frame(
            header=urwid.Pile([
                display_header,
                QuestionTitle(self.question_title),
                self.question_text,
                QuestionStats(self.question_stats),
                urwid.Divider('-')
            ]),
            body=self.answer_text,
            footer=urwid.Pile([
                QuestionURL(self.url),
                UnicodeText(
                    'o: open in browser, v: back to answer, \u2190: back, q: quit')
            ])
        )
        return comment_frame

    def keypress(self, size, key):
        """
        Overrides keypress in superclass, so don't fall for the trap! size parameter is needed!
        """
        if key in {'down', 'n', 'N'} and not self.answer_text.comments_toggled:
            # bool comparison is necessary to disable up down buttons when comments are being shown
            self.answer_text.next_ans()
        elif key in {'up', 'b', 'B'} and not self.answer_text.comments_toggled:
            self.answer_text.prev_ans()
        elif key in {'c', 'C'}:
            self.answer_text.show_comments()
            self._invalidate()
            comment_frame = self.make_comment_frame()
            urwid.WidgetWrap.__init__(self, comment_frame)
            self.answer_text.comments_toggled = True
        elif key in {'v', 'V'}:
            self.answer_text.set_content()
            self._invalidate()
            answer_frame = self.make_frame()
            urwid.WidgetWrap.__init__(self, answer_frame)
            self.answer_text.comments_toggled = False
        elif key in {'o', 'O'}:
            import webbrowser
            display_header.event('browser', "Opening in your browser...")
            webbrowser.open(self.url)
        elif key == 'left':
            global question_post
            global question_page
            question_post = None
            if question_page is None:
                sys.exit(0)
            else:
                MAIN_LOOP.widget = question_page
        elif key == 'window resize':
            screen_height, screen_width = subprocess.check_output(
                ['stty', 'size']).split()
            if self.screenHeight != screen_height and not self.answer_text.comments_toggled:
                self._invalidate()
                answer_frame = self.make_frame()
                urwid.WidgetWrap.__init__(self, answer_frame)
            elif self.screenHeight != screen_height and self.answer_text.comments_toggled:
                self._invalidate()
                comment_frame = self.make_comment_frame()
                urwid.WidgetWrap.__init__(self, comment_frame)
        elif key in {'q', 'Q'}:
            sys.exit(0)
        elif key in {'d', 'D'}:
            if self.dup_url:
                pr.display_results(self.dup_url, self.url)
            elif self.dup_link:
                pr.display_results(self.dup_link) 


class AnswerText(urwid.WidgetWrap):
    """Answers to the question.

    Long answers can be navigated up or down using the mouse.
    """

    def __init__(self, answers, comments):
        urwid.WidgetWrap.__init__(self, UnicodeText(''))
        self._selectable = True  # so that we receive keyboard input
        self.answers = answers
        self.comments_list = comments
        # if the comments are being shown then comments_toggled will be True else when are answers are being
        # shown then comments_toggled will be False
        # This Bool is necessary to disable up/down arrow keys when comments are being shown
        self.comments_toggled = False
        self.index = 0
        self.set_content()

    def set_content(self):
        """
        We must use a box adapter to get the text to scroll when this widget is already in
        a Pile from the main question page. Scrolling is necessary for long answers which are longer
        than the length of the terminal.
        """
        self.content = [('less-important', 'Answer: ')] + self.answers[self.index].split("\n")
        self._w = ScrollableTextBox(self.content)

    def prev_ans(self):
        """go to previous answer."""
        self.index -= 1
        if self.index < 0:
            self.index = 0
            display_header.event('answer-bounds', "No previous answers.")
        else:
            display_header.clear('answer-bounds')
        self.set_content()

    def next_ans(self):
        """go to next answer."""
        self.index += 1
        if self.index > len(self.answers) - 1:
            self.index = len(self.answers) - 1
            display_header.event('answer-bounds', "No more answers.")
        else:
            display_header.clear('answer-bounds')
        self.set_content()

    def show_comments(self):
        """Shows comments by loading a new frame name QuestionPage.make_comment_frame()"""
        self.content = [('less-important', 'Comments: \n')] + \
                       self.comments_list[self.index]
        self._w = ScrollableTextBox(self.content)

    def __len__(self):
        """ return number of rows in this widget """
        return len(self.content)


class ScrollableTextBox(urwid.ListBox):
    """ Display input text, scrolling through when there is not enough room.

    Scrolling through text takes a little work to support on Urwid.
    """

    def __init__(self, content):
        """
        :param content: text string to be displayed
        """
        lines = [UnicodeText(line) for line in content]
        body = urwid.SimpleFocusListWalker(lines)
        urwid.ListBox.__init__(self, body)

    def mouse_event(self, size, event, button, col, row, focus):
        SCROLL_WHEEL_UP = 4
        SCROLL_WHEEL_DOWN = 5
        if button == SCROLL_WHEEL_DOWN:
            self.keypress(size, 'down')
        elif button == SCROLL_WHEEL_UP:
            self.keypress(size, 'up')
        else:
            return False
        return True


class QuestionTitle(UnicodeText):
    """ Title of the question,"""

    def __init__(self, title):
        text = ["Question: ", ('title', title), "\n"]
        UnicodeText.__init__(self, text)


# Must convert to BoxAdapter object if used as a flow widget.
class QuestionDescription(urwid.WidgetWrap):
    """ Description of the question """

    def __init__(self, description):
        urwid.WidgetWrap.__init__(self, UnicodeText(''))
        self.description = description
        self.set_description()

    def set_description(self):
        """
        We must use a box adapter to get the text to scroll when this widget is already in
        a Pile from the main question page. Scrolling is necessary for long questions which are longer
        than the length of the terminal.
        """
        self.content = self.description.strip("\n").split("\n")
        self._w = ScrollableTextBox(self.content)

    def __len__(self):
        """ return number of rows in this widget """
        return len(self.content)


class QuestionStats(UnicodeText):
    """ Stats of the question,"""

    def __init__(self, stats):
        text = ["\n", ('metadata', stats)]
        UnicodeText.__init__(self, text)


class QuestionURL(UnicodeText):
    """ url of the question """

    def __init__(self, url):
        text = ["\n", ('heading', 'Question URL: '), url]
        UnicodeText.__init__(self, text)



"""
Tests for socli
"""

import sys
import socli.socli as _socli
import socli.user as _user
import socli.search as _search

sys.path.append("..")


squery = "python for loop"
surl = "https://stackoverflow.com/questions/522563/accessing-the-index-in-for-loops?rq=1"

def test_wrongsyn():
    """ socli.wrongsyn() detects `"sdf"` as non-empty string """
    assert _socli.wrongsyn("sdf") == None

def test_fix_google_url():
    """ socli.fix_google_url() adds `https` protocol if not existent yet """
    url = _search.fix_google_url("www.example.com/questions/1234")
    assert url == "https://www.example.com/questions/1234"

def load_user_agents():
    _search.load_user_agents()

def test_parse_query():
    command = "python for loop".split()
    namespace = _socli.parse_arguments(command)

    assert not namespace.api
    assert namespace.browse == []
    assert not namespace.debug
    assert not namespace.delete
    assert not namespace.help
    assert not namespace.interactive
    assert not namespace.new
    assert namespace.query == []
    assert namespace.userQuery == "python for loop".split()
    assert not namespace.res
    assert not namespace.sosearch
    assert not namespace.tag
    assert not namespace.user
    assert not namespace.open_url
    assert not namespace.version

def test_parse_help():
    command = "-h".split()
    namespace = _socli.parse_arguments(command)

    assert not namespace.api
    assert namespace.browse == []
    assert not namespace.debug
    assert not namespace.delete
    assert namespace.help
    assert not namespace.interactive
    assert not namespace.new
    assert namespace.query == []
    assert not namespace.res
    assert not namespace.sosearch
    assert not namespace.tag
    assert not namespace.user
    assert not namespace.open_url
    assert not namespace.version

def test_parse_version():
    command = "-v".split()
    namespace = _socli.parse_arguments(command)

    assert not namespace.api
    assert namespace.browse == []
    assert not namespace.debug
    assert not namespace.delete
    assert not namespace.help
    assert not namespace.interactive
    assert not namespace.new
    assert namespace.query == []
    assert not namespace.res
    assert not namespace.sosearch
    assert not namespace.tag
    assert not namespace.user
    assert not namespace.open_url
    assert namespace.version

def test_parse_interactive():
    command = "-iq python for loop".split()
    namespace = _socli.parse_arguments(command)

    assert not namespace.api
    assert namespace.browse == []
    assert not namespace.debug
    assert not namespace.delete
    assert not namespace.help
    assert namespace.interactive
    assert not namespace.new
    assert namespace.query == "python for loop".split()
    assert not namespace.res
    assert not namespace.sosearch
    assert not namespace.tag
    assert not namespace.user
    assert not namespace.open_url
    assert not namespace.version

def test_parse_open_url():
    command = "--open-url https://stackoverflow.com/questions/20639180/explanation-of-how-nested-list-comprehension-works ".split()
    namespace = _socli.parse_arguments(command)

    assert not namespace.api
    assert namespace.browse == []
    assert not namespace.debug
    assert not namespace.delete
    assert not namespace.help
    assert not namespace.interactive
    assert not namespace.new
    assert namespace.query == []
    assert not namespace.res
    assert not namespace.sosearch
    assert not namespace.tag
    assert not namespace.user
    assert namespace.open_url!=[]
    assert not namespace.version

def test_user_json():
    try:
        _user.app_data = {"user": "John Smith"}
        _user.save_datafile()
        _user.load_datafile()
        _user.del_datafile()
    except Exception:
        raise SoCLITestingException("User JSON test failed.")

    try:
        name = "John Smith"
        _user.app_data = {"user": name}
        _user.save_datafile()
        user = _user.retrieve_saved_profile()
        assert user == name
        _user.del_datafile()
    except Exception:
        raise SoCLITestingException("User JSON test failed.")

# Commented out because of stackoverflow bot detection
# def test_searchSO():
#     try:
#         _search.google_search = False
#         _search.get_questions_for_query(squery)
#     except Exception:
#         raise SoCLITestingException("Search Stack Overflow test failed.")


def test_searchStats():
    try:
        _search.get_question_stats_and_answer_and_comments(surl)
    except Exception:
        raise SoCLITestingException("Search SO stats test failed.")


class SoCLITestingException(Exception):

    def __init__(self, msg):
        super().__init__(msg)



"""
Contains all functions used for printing.
Uses colorama for formatting.
"""

import subprocess
import sys
import textwrap
import urllib
import json

import colorama
import requests

from socli import search, tui

DEBUG = False

# Bold and underline are not supported by colorama.
_BOLD = '\033[1m'
_UNDERLINE = '\033[4m'

palette = [('answer', 'default', 'default'),
           ('title', 'light green, bold', 'default'),
           ('heading', 'light green, bold', 'default'),
           ('metadata', 'dark green', 'default'),
           ('less-important', 'dark gray', 'default'),
           ('warning', 'yellow', 'default')
           ]

if sys.version < '3.0.0':
    global FileNotFoundError
    FileNotFoundError = IOError

    def urlencode(inp):
        return urllib.quote_plus(inp)

    def display_str(inp):
        return inp.encode('utf-8')

    def inputs(str=""):
        sys.stdout.write(str)
        tempx = raw_input()
        return tempx
else:
    def urlencode(inp):
        return urllib.parse.quote_plus(inp)

    def display_str(inp):
        return inp

    def inputs(string=""):
        sys.stdout.write(string)
        tempx = input()
        return tempx


def format_str(string, color):
    return "{0}{1}{2}".format(color, string, colorama.Style.RESET_ALL)


def print_header(string):
    print(format_str(string, colorama.Fore.MAGENTA))


def print_blue(string):
    print(format_str(string, colorama.Fore.BLUE))


def print_green(string):
    print(format_str(string, colorama.Fore.GREEN))


def print_warning(string):
    print(format_str(string, colorama.Fore.YELLOW))


def print_fail(string):
    print(format_str(string, colorama.Fore.RED))


def print_white(string):
    print(format_str(string, colorama.Fore.WHITE))


def make_header(string):
    return format_str(string, colorama.Fore.MAGENTA)


def make_blue(string):
    return format_str(string, colorama.Fore.BLUE)


def make_green(string):
    return format_str(string, colorama.Fore.GREEN)


def make_warning(string):
    return format_str(string, colorama.Fore.YELLOW)


def make_fail(string):
    return format_str(string, colorama.Fore.RED)


def make_white(string):
    return format_str(string, colorama.Fore.WHITE)


def bold(string):
    return format_str(string, _BOLD)


def underline(string):
    return format_str(string, _UNDERLINE)


# For testing exceptions
def showerror(e):
    if DEBUG:
        import traceback
        print("Error name: " + e.__doc__)
        print()
        print("Description: " + str(e))
        print()
        traceback.print_exc()
    else:
        return


def helpman():
    """
    Displays help
    :return:
    """

    options_text = make_header("Stack Overflow command line client:") + '\n' + \
                   make_green("\n\n\tUsage: socli [ Arguments ] < Search Query >\n\n") + '\n' + \
                   make_header("[ Arguments ] (optional):\n") + '\n' + \
                   " " + bold("--help or -h") + " : Displays this help" + '\n' + \
                   " " + bold("--query or -q") + \
                   " : If any of the following commands are used then you " \
                   "must specify search query after the query argument" + '\n' + \
                   " " + bold("--interactive or -i") + " : To search in Stack Overflow" + \
                   " and display the matching results. You can chose and " + \
                   "browse any of the result interactively" + '\n' + \
                   " " + bold("--res or -r") + \
                   " : To select and display a result manually and display " + \
                   "its most voted answer. \n    eg: " + \
                   make_warning(
                       "socli --res 2 --query foo bar") + ": Displays the second search result of the query" + \
                   " \"foo bar\"" + '\n' + \
                   " " + bold("--tag or -t") + \
                   " : To search a query by tag on Stack Overflow.  Visit http://stackoverflow.com/tags to see the " + \
                   "list of all tags." + \
                   "\n    eg: " + make_warning("socli --tag javascript,node.js --query foo bar") + \
                   ": Displays the search result of the query" + \
                   " \"foo bar\" in Stack Overflow's javascript and node.js tags." + '\n' + \
                   " " + bold("--new or -n") + \
                   " : Opens the Stack Overflow new questions page in your default browser. You can create a " + \
                   "new question using it." + '\n' + \
                   " " + bold("--user or -u") + \
                   " : Displays information about the user provided as the next argument(optional). " \
                   "If no argument is provided it will ask the user to enter a default username. " \
                   "Now the user can run the command without the argument." + \
                   "\n    eg: " + make_warning(
        "socli -u") + ": Prompts and saves your username. Now you can just run " + \
                   make_warning("socli -u") + " to see " + \
                   "the stats.\n    " + make_warning(
        "socli -u 22656") + ": Displays info about user ID 22656" + '\n' + \
                   " " + bold("--del or -d") + \
                   " : Deletes the configuration file generated by " + make_warning("socli -u") + " command." + '\n' + \
                   " " + bold("--api or -a") + \
                   " : Sets a custom API key for socli" + '\n' + \
                   " " + bold("--sosearch or -s") + \
                   " : SoCLI uses google search by default. Use this option to search Stack Overflow directly." + '\n' \
                   " " + bold("--open-url or -o") + \
                   " : Opens the given url in socli " +  '\n' \
                   " " + bold("--version or -v") + \
                   " : Displays the current version of socli " +  '\n'

    help_text = make_header("\n\n< Search Query >:") + '\n' + \
               "\nQuery to search on Stack Overflow" + '\n' + \
               "\nIf no commands are specified then socli will search Stack " + \
               "Overflow and simply displays the first search result's " + \
               "most voted answer." + '\n' + \
               "If a command is specified then it will work according to the " + \
               "command." + '\n' + \
               make_header("\n\nExamples:\n") + '\n' + \
               '\t' + make_warning("socli for loop in python") + '\n' + \
               '\t' + make_warning("socli -iq while loop in python") + '\n' + \
               "\n\nSoCLI is an open source project hosted on github. Don't forget to star it if you liked it.\nUse GitHub" + \
               " issues to report problems: " + underline("http://github.com/gautamkrishnar/socli")

    screen_height, screen_width = subprocess.check_output(['stty', 'size']).split()
    subsequent_indent = '    '
    options_text = '\n'.join(['\n'.join(textwrap.wrap(line, width=int(screen_width) - len(subsequent_indent),
                                                      break_long_words=False, replace_whitespace=False,
                                                      subsequent_indent=subsequent_indent))
                              for line in options_text.splitlines() if options_text.strip() != ''])

    help_text = '\n'.join(['\n'.join(textwrap.wrap(line, width=int(screen_width),
                                                  break_long_words=False, replace_whitespace=False))
                          for line in help_text.splitlines() if help_text.strip() != ''])
    print(options_text)
    print(help_text)


def display_results(url, dup_link=None, json_output=False):
    """
    Display result page
    :param url: URL of the search result
    :param dup_link: URL to the duplicate question visited from
    :param json_output: JSON output flag
    :return:
    """
    search.random_headers()
    res_page = requests.get(url, headers=search.header)
    search.captcha_check(res_page.url)
    question_title, question_desc, question_stats, answers, comments, dup_url = \
        search.get_question_stats_and_answer_and_comments(url)
    if json_output:
        sys.stdout.write(urllib.parse.unquote(json.dumps({
            'title': question_title,
            'desc': question_desc,
            'stats': question_stats,
            'answers': answers,
            'comments': comments,
            'dup_url': dup_url,
        })))
    else:
        tui.display_header = tui.Header()
        tui.question_post = tui.QuestionPage((url, question_title, question_desc, question_stats, answers, comments, dup_url, dup_link))
        tui.MAIN_LOOP = tui.EditedMainLoop(tui.question_post, palette)
        tui.MAIN_LOOP.run()



"""
# Stack Overflow CLI
# Created by
# Gautam Krishna R : www.github.com/gautamkrishnar
# And open source contributors at GitHub: https://github.com/gautamkrishnar/socli#contributors
"""

import os
import re
import sys
import logging
import requests
from bs4 import BeautifulSoup
import urwid
import urllib3

try:
    import simplejson as json
except ImportError:
    import json
try:
    JSONDecodeError = json.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError

# Importing SoCli modules
import socli.tui as tui
import socli.user as user_module
import socli.search as search
import socli.printer as printer
from socli.parser import parse_arguments
from socli.printer import display_results
from socli.version import __version__

tag = ""  # tag based search
query = ""  # Query

# Suppressing InsecureRequestWarning and many others
urllib3.disable_warnings()
# logger for debugging
logger = logging.getLogger(__name__)


# Switch on logging of the requests module.
def debug_requests_on():
    try:
        from http.client import HTTPConnection
        HTTPConnection.set_debuglevel(HTTPConnection, 1)
    except ImportError:
        import httplib
        httplib.HTTPConnection.debuglevel = 2

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger('requests.packages.urllib3')
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


# Fixes windows active page code errors
def fix_code_page():
    if sys.platform == 'win32':
        if sys.stdout.encoding != 'cp65001':
            os.system("echo off")
            os.system("chcp 65001")  # Change active page code
            sys.stdout.write("\x1b[A")  # Removes the output of chcp command
            sys.stdout.flush()
            return
        return


def wrongsyn(query):
    """
    Exits if query value is empty
    :param query:
    :return:
    """
    if not query:
        printer.print_warning("Wrong syntax!...\n")
        printer.helpman()
        sys.exit(1)
    else:
        return


def has_tags():
    """
    Gets the tags and adds them to query url
    :return:
    """
    global tag
    for tags in tag:
        search.so_qurl = search.so_qurl + "[" + tags + "]" + "+"


def socli(query, json_output=False):
    """
    SoCLI Code
    :param query: Query to search on Stack Overflow.
    If google_search is true uses Google search to find the best result.
    Else, use Stack Overflow default search mechanism.
    :return:
    """
    query = printer.urlencode(query)
    try:
        if search.google_search:
            questions = search.get_questions_for_query_google(query)
            res_url = questions[0][2]  # Gets the first result
            display_results(res_url, json_output=json_output)
        else:
            questions = search.get_questions_for_query(query)
            res_url = questions[0][2]
            display_results(search.so_url + res_url, json_output=json_output)  # Returned URL is relative to SO homepage

    except IndexError:
        if len(questions) > 1:
            for i in range(1, len(questions)):
                res_url = questions[i][2]
                try:
                    display_results(res_url, json_output=json_output)
                except IndexError:
                    continue
                break
        printer.print_warning("No results found...")

    except UnicodeEncodeError as e:
        printer.showerror(e)
        printer.print_warning("\n\nEncoding error: Use \"chcp 65001\" command before using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        printer.showerror(e)
        sys.exit(0)


def socli_browse_interactive_windows(query_tag):
    """
    Interactive mode for -b browse
    :param query_tag:
    :return:
    """
    try:
        search_res = requests.get(search.so_burl + query_tag)
        search.captcha_check(search_res.url)
        soup = BeautifulSoup(search_res.text, 'html.parser')
        try:
            soup.find_all("div", class_="question-summary")[0]  # For explicitly raising exception
            tmp = (soup.find_all("div", class_="question-summary"))
            i = 0
            question_local_url = []
            print(printer.bold("\nSelect a question below:\n"))
            while i < len(tmp):
                if i == 10:
                    break  # limiting results
                question_text = ' '.join((tmp[i].a.get_text()).split())
                question_text = question_text.replace("Q: ", "")
                printer.print_warning(str(i + 1) + ". " + printer.display_str(question_text))
                q_tag = (soup.find_all("div", class_="question-summary"))[i]
                answers = [s.get_text() for s in q_tag.find_all("a", class_="post-tag")][0:]
                ques_tags = " ".join(str(x) for x in answers)
                question_local_url.append(tmp[i].a.get("href"))
                print("  " + printer.display_str(ques_tags) + "\n")
                i = i + 1
            try:
                op = int(printer.inputs("\nType the option no to continue or any other key to exit:"))
                while 1:
                    if (op > 0) and (op <= i):
                        display_results(search.so_burl + question_local_url[op - 1])
                        cnt = 1  # this is because the 1st post is the question itself
                        while 1:
                            global tmpsoup
                            qna = printer.inputs(
                                "Type " + printer.bold("o") + " to open in browser, " + printer.bold(
                                    "n") + " to next answer, " + printer.bold(
                                    "b") + " for previous answer or any other key to exit:")
                            if qna in ["n", "N"]:
                                try:
                                    answer = (tmpsoup.find_all("div", class_="js-post-body")[cnt + 1].get_text())
                                    printer.print_green("\n\nAnswer:\n")
                                    print("-------\n" + answer + "\n-------\n")
                                    cnt = cnt + 1
                                except IndexError:
                                    printer.print_warning(" No more answers found for this question. Exiting...")
                                    sys.exit(0)
                                continue
                            if qna in ["b", "B"]:
                                if cnt == 1:
                                    printer.print_warning(" You cant go further back. You are on the first answer!")
                                    continue
                                answer = (tmpsoup.find_all("div", class_="js-post-body")[cnt - 1].get_text())
                                printer.print_green("\n\nAnswer:\n")
                                print("-------\n" + answer + "\n-------\n")
                                cnt = cnt - 1
                                continue
                            if qna in ["o", "O"]:
                                import webbrowser
                                printer.print_warning("Opening in your browser...")
                                webbrowser.open(search.so_burl + question_local_url[op - 1])
                            else:
                                break
                        sys.exit(0)
                    else:
                        op = int(input("\n\nWrong option. select the option no to continue:"))
            except Exception as e:
                printer.showerror(e)
                printer.print_warning("\n Exiting...")
                sys.exit(0)
        except IndexError:
            printer.print_warning("No results found...")
            sys.exit(0)

    except UnicodeEncodeError:
        printer.print_warning("\n\nEncoding error: Use \"chcp 65001\" command before using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        printer.showerror(e)
        sys.exit(0)


def socli_browse_interactive(query_tag):
    """
    Interactive mode
    :return:
    """
    if sys.platform == 'win32':
        return socli_browse_interactive_windows(query_tag)

    class SelectQuestionPage(urwid.WidgetWrap):

        def display_text(self, index, question):
            question_text, question_desc, _ = question
            text = [
                ("warning", u"{}. {}\n".format(index, question_text)),
                question_desc + "\n",
            ]
            return text

        def __init__(self, questions):
            self.questions = questions
            self.cachedQuestions = [None for _ in range(10)]
            widgets = [self.display_text(i, q) for i, q in enumerate(questions)]
            self.questions_box = tui.ScrollableTextBox(widgets)
            self.header = tui.UnicodeText(('less-important', 'Select a question below:\n'))
            self.footerText = '0-' + str(len(self.questions) - 1) + ': select a question, any other key: exit.'
            self.errorText = tui.UnicodeText.to_unicode('Question numbers range from 0-' +
                                                        str(len(self.questions) - 1) +
                                                        ". Please select a valid question number.")
            self.footer = tui.UnicodeText(self.footerText)
            self.footerText = tui.UnicodeText.to_unicode(self.footerText)
            frame = urwid.Frame(header=self.header,
                                body=urwid.Filler(self.questions_box, height=('relative', 100), valign='top'),
                                footer=self.footer)
            urwid.WidgetWrap.__init__(self, frame)

        # Override parent method
        def selectable(self):
            return True

        def keypress(self, size, key):
            if key in '0123456789':
                try:
                    question_url = self.questions[int(key)][2]
                    self.footer.set_text(self.footerText)
                    self.select_question(question_url, int(key))
                except IndexError:
                    self.footer.set_text(self.errorText)
            elif key in {'down', 'up'}:
                self.questions_box.keypress(size, key)
            else:
                raise urwid.ExitMainLoop()

        def select_question(self, url, index):
            if self.cachedQuestions[index] is not None:
                tui.question_post = self.cachedQuestions[index]
                tui.MAIN_LOOP.widget = tui.question_post
            else:
                if not search.google_search:
                    url = search.so_url + url
                question_title, question_desc, question_stats, answers, comments, dup_url = \
                    search.get_question_stats_and_answer_and_comments(url)
                question_post = tui.QuestionPage(
                    (url, question_title, question_desc, question_stats, answers, comments, dup_url, None)
                )
                self.cachedQuestions[index] = question_post
                tui.MAIN_LOOP.widget = question_post

    tui.display_header = tui.Header()

    try:
        if search.google_search:
            questions = search.get_questions_for_query_google(query)
        else:
            # print('hurr')
            questions = search.get_questions_for_query(query_tag)
            # print(questions)

        question_page = SelectQuestionPage(questions)

        tui.MAIN_LOOP = tui.EditedMainLoop(question_page, printer.palette)
        tui.MAIN_LOOP.run()

    except UnicodeEncodeError:
        printer.print_warning("\n\nEncoding error: Use \"chcp 65001\" command before using socli...")
        sys.exit(0)
    except requests.exceptions.ConnectionError:
        printer.print_fail("Please check your internet connectivity...")
    except Exception as e:
        printer.showerror(e)
        # print("Hurra")
        print("exiting...")
        sys.exit(0)


def main():
    """
    The main logic for how options in a command is checked.
    """
    global query
    namespace = parse_arguments(sys.argv[1:])
    search.load_user_agents()  # Populates the user agents array
    query_tag = ' '.join(namespace.browse)  # Tags

    # Query
    if namespace.query:
        # Query when args are present
        query = ' '.join(namespace.query)
    elif namespace.userQuery:
        # Query without any args
        query = ' '.join(namespace.userQuery)

    if namespace.help:
        # Display command line syntax
        printer.helpman()
        sys.exit(0)

    json_output = bool(namespace.json)

    if namespace.debug:  # If --debug flag is present
        # Prints out error used for debugging
        printer.DEBUG = True
        debug_requests_on()

    if namespace.register:
        os.system('eval "$(register-python-argcomplete socli)"')
        sys.exit(0)

    if namespace.new:  # If --new flag is present
        # Opens StackOverflow website in the browser to create a  new question
        import webbrowser
        printer.print_warning("Opening stack overflow in your browser...")
        webbrowser.open(search.so_url + "/questions/ask")
        sys.exit(0)

    if namespace.api:  # If --api flag is present
        # Sets custom API key
        user_module.set_api_key()
        sys.exit(0)

    if namespace.user is not None:  # If --user flag is present
        # Stackoverflow user profile support
        if namespace.user != '(RANDOM_STRING_CONSTANT)':  # If user provided a user ID
            user_module.manual = 1  # Enabling manual mode
            user = namespace.user
        else:  # If user did not provide a user id
            user = user_module.retrieve_saved_profile()  # Reading saved user id from app data
        user_module.user_page(user)
        sys.exit(0)

    if namespace.delete:  # If --delete flag is present
        # Deletes user data
        user_module.del_datafile()
        printer.print_warning("Data files deleted...")
        sys.exit(0)

    if namespace.sosearch:  # If --sosearch flag is present
        # Disables google search
        search.google_search = False

    if namespace.tag:  # If --tag flag is present
        global tag
        search.google_search = False
        tag = namespace.tag
        has_tags()  # Adds tags to StackOverflow url (when not using google search.
    if namespace.open_url:
        import webbrowser
        open_in_browser = False
        display_condition = True
        url_to_use = namespace.open_url[0]
        if re.findall(r"^https:\/\/", url_to_use):
            pass
        else:
            url_to_use = "https://" + url_to_use
        try:
            if url_to_use == "https://stackoverflow.com/questions/":
                raise Exception('URL Error')
            if url_to_use == "https://www.stackoverflow.com/questions/":
                raise Exception('URL Error')
            requests.get(url_to_use)
        except Exception:
            printer.print_warning(
                "Error, could be:\n- invalid url\n- url cannot be opened in socli\n- internet connection error")
            sys.exit(0)
        nostackoverflow = re.findall(r"stackoverflow\.com", url_to_use)
        if not nostackoverflow:
            open_in_browser = True
            display_condition = False
            printer.print_warning("Your url is not a stack overflow url.\nOpening in your browser...")
        tag_matcher = re.findall(r"\/tag.+\/", url_to_use)
        blog_matcher = re.findall(r"blog", url_to_use)
        if tag_matcher:
            extracted_tag = ""
            if not re.findall(r"tagged", url_to_use):
                extracted_tag = re.split(r"\/", url_to_use)[4]
            else:
                extracted_tag = re.split(r"\/", url_to_use)[5]
            open_in_browser = False
            display_condition = False
            tag = extracted_tag
            search.socli_interactive(tag)
        if blog_matcher:
            open_in_browser = True
            display_condition = False
            printer.print_warning("Your url belongs to blog")
            printer.print_warning("Opening in browser...")
        if display_condition:
            open_in_browser = False
            try:
                display_results(url_to_use, json_output=json_output)
            except IndexError:
                printer.print_fail("The URL specified is returning a 404, please check the url and try again!")
                sys.exit(0)
        if open_in_browser:
            webbrowser.open(url_to_use)
    if namespace.res is not None:  # If --res flag is present
        # Automatically displays the result specified by the number
        question_number = namespace.res
        if namespace.query != [] or namespace.tag is not None:  # There must either be a tag or a query
            search.socli_manual_search(query, question_number)
        else:
            printer.print_warning(
                'You must specify a query or a tag. For example, use: "socli -r 3 -q python for loop" '
                'to retrieve the third result when searching about "python for loop". '
                'You can also use "socli -r 3 -t python" '
                'to retrieve the third result when searching for posts with the "python" tag.')


    if namespace.version:
        print('Socli ' + __version__)
        sys.exit(0)

    if namespace.browse:
        # Browse mode
        search.google_search = False
        socli_browse_interactive(query_tag)
    elif namespace.query != [] or namespace.tag is not None:  # If query and tag are not both empty
        if namespace.interactive:
            search.socli_interactive(query)
        else:
            socli(query, json_output=json_output)
    elif query not in [' ', ''] and not (
            namespace.tag or namespace.res or namespace.interactive or namespace.browse):  # If there are no flags
        socli(query, json_output=json_output)
    else:
        # Help text for interactive mode
        if namespace.interactive and namespace.query == [] and namespace.tag is None:
            printer.print_warning('You must specify a query or a tag. For example, use: "socli -iq python for loop" '
                                  'to search about "python for loop" in interactive mode. '
                                  'You can also use "socli -it python" '
                                  'to search posts with the "python" tag in interactive mode.')
        else:
            printer.helpman()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from codecs import open
from distutils.util import convert_path

with open('README.md') as f:
    longd = f.read()

main_ns = {}
ver_path = convert_path('socli/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name='socli',
    include_package_data=True,
    packages=["socli"],
    entry_points={"console_scripts": ['socli = socli.sentry:main']},
    python_requires='>=3.5.0',
    install_requires=['BeautifulSoup4', 'requests', 'colorama', 'Py-stackExchange', 'sentry_sdk', 'urwid', 'argcomplete'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    requires=['BeautifulSoup4', 'requests', 'colorama', 'PyStackExchange', 'sentry_sdk', 'urwid'],
    data_files=[('man/man1',['socli.1'])],
    version=main_ns['__version__'],
    url='https://www.github.com/gautamkrishnar/socli',
    keywords="stack overflow cli",
    license='BSD',
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    author='Gautam Krishna R',
    author_email='r.gautamkrishna@gmail.com',
    description='Stack overflow command line interface. SoCLI allows you to search and browse stack overflow from the '
                'terminal.',
    long_description="\n\n" + longd,
    long_description_content_type='text/markdown'
)



"""
Contains all functions used for user authentication and user metadata fetching.
"""

import json
import os
import sys
import socli.printer as pr

app_data = {} # Data file dictionary
data_file = os.path.join(os.path.dirname(__file__), "data.json")  # Data file location


def user_page(user_id):
    """
    Stack Overflow user profile browsing
    :param user_id:
    :return:
    """
    global app_data
    import stackexchange

    try:
        user_id = int(user_id)
    except ValueError:
        pr.print_warning("\nUser ID must be an integer.")
        print(
            "\nFollow the instructions on this page to get your User ID: http://meta.stackexchange.com/a/111130")
        sys.exit()

    try:
        from urllib.error import URLError
    except ImportError:
        from urllib import URLError
    try:
        user_id = int(user_id)
    except ValueError:
        pr.print_warning("\nUser ID must be an integer.")
        print(
            "\nFollow the instructions on this page to get your User ID: http://meta.stackexchange.com/a/111130")
        sys.exit()

    try:
        if "api_key" not in app_data:
            app_data["api_key"] = None
        userprofile = stackexchange.Site(stackexchange.StackOverflow, app_key=app_data["api_key"]).user(user_id)
        print(pr.bold("\n User: " + userprofile.display_name.format()))
        print("\n\tReputations: " + userprofile.reputation.format())
        pr.print_warning("\n\tBadges:")
        print("\t\t   Gold: " + str(userprofile.gold_badges))
        print("\t\t Silver: " + str(userprofile.silver_badges))
        print("\t\t Bronze: " + str(userprofile.bronze_badges))
        print("\t\t  Total: " + str(userprofile.badge_total))
        pr.print_warning("\n\tStats:")
        total_questions = len(userprofile.questions.fetch())
        unaccepted_questions = len(userprofile.unaccepted_questions.fetch())
        accepted = total_questions - unaccepted_questions
        print("\t\t Total Questions Asked: " + str(len(userprofile.questions.fetch())))
        try:
            # the following division might raise ZeroDivisionError if
            # total_questions is zero, so we wrap it in try-except.
            rate = accepted / float(total_questions) * 100
            print('\t\t        Accept rate is: %.2f%%.' % rate)
        except ZeroDivisionError:
            pass  # if total_question is zero we don't print accept rate.

        # check if the user have answers and questions
        if userprofile.top_answer_tags.fetch():
            print('\nMost experienced on %s.' % userprofile.top_answer_tags.fetch()[0].tag_name)
        else:
            print("You have 0 answers")
        if userprofile.top_question_tags.fetch():
            print('Most curious about %s.' % userprofile.top_question_tags.fetch()[0].tag_name)
        else:
            print("You have 0 questions")
    except URLError:
        pr.print_fail("Please check your internet connectivity...")
        sys.exit()
    except Exception as e:
        pr.showerror(e)
        if str(e) == "400 [bad_parameter]: `key` doesn't match a known application":
            pr.print_warning("Wrong API key... Deleting the data file...")
            del_datafile()
            sys.exit()
        elif str(e) in ("not enough values to unpack (expected 1, got 0)", "400 [bad_parameter]: ids"):
            global manual
            if manual:
                pr.print_warning("Wrong user ID specified...")
                pr.helpman()
                sys.exit()
            pr.print_warning("Wrong user ID... Deleting the data file...")
            del_datafile()
            sys.exit()

        # Reaches here when rate limit exceeds
        pr.print_warning(
            "Stack Overflow exception. This might be caused due to the rate limiting: "
            "http://stackapps.com/questions/3055/is-there-a-limit-of-api-requests")
        print("Use http://stackapps.com/apps/oauth/register to register a new API key.")
        set_api_key()
        sys.exit()


def set_api_key():
    """
    Sets a custom API Key
    :return:
    """
    global app_data
    try:
        api_key = pr.inputs("Type an API key to continue (^C to abort): ")
        if len(api_key) > 0:
            app_data["api_key"] = api_key
            save_datafile()
        pr.print_warning("\nAPI Key saved...")
    except KeyboardInterrupt:
        print("Aborted.")


def save_datafile():
    """
    Saves the app_data dictionary to a file named data_file
    :return:
    """
    # import json => Json imported globally
    global app_data
    global data_file
    with open(data_file, "w") as dataf:
        json.dump(app_data, dataf)


def load_datafile():
    """
    Loads the app_data dictionary form a file named data_file
    :return:
    """
    # import json => Json imported globally
    global app_data
    global data_file
    with open(data_file) as dataf:
        app_data = json.load(dataf)


def del_datafile():
    """
    Deletes the data file
    :return:
    """
    global data_file
    try:
        os.remove(data_file)
    except FileNotFoundError:
        pr.print_warning("File not created.... Use socli -u to create a new configuration file.")
        exit(0)


def retrieve_saved_profile():
    """
    Retrieves the user's saved profile after a "socli -u" command.
    Asks the user to enter a User ID and saves it if a previous file is not found.

    :return: The user's ID as an integer
    """
    global data_file
    global app_data
    user = None
    try:
        load_datafile()
        if "user" in app_data:
            user = app_data["user"]
        else:
            raise FileNotFoundError  # Manually raising to get value
    except json.JSONDecodeError:
        # This maybe some write failures
        del_datafile()
        pr.print_warning("Error in parsing the data file, it will be now deleted. Please rerun the "
                         "socli -u command.")
        sys.exit()
    except FileNotFoundError:
        pr.print_warning("Default user not set...\n")
        try:
            # Code to execute when first time user runs socli -u
            app_data['user'] = int(pr.inputs("Enter your Stackoverflow User ID: "))
            save_datafile()
            user = app_data['user']
            pr.print_green("\nUserID saved...\n")
        except ValueError:
            pr.print_warning("\nUser ID must be an integer.")
            print(
                "\nFollow the instructions on this page to get your User ID: http://meta.stackexchange.com/a/111130")
            sys.exit()
    return user



"""
Contains all functions used for socli authentication
"""
import os
from datetime import datetime

from functools import wraps
from getpass import getpass

from bs4 import BeautifulSoup
from requests import Session

# Supporting input in Python 2/3
try:
    input = raw_input
except NameError:
    pass

# Supporting LWPCookieJar in Python 2/3
try:
    from http.cookiejar import LWPCookieJar
except ImportError:
    from cookielib import LWPCookieJar

COOKIES_FILE_PATH = '.cookies'
BASE_URL = 'https://stackoverflow.com/'
LOGIN_URL = BASE_URL + 'users/login'
LOGOUT_URL = BASE_URL + 'users/logout'


def login_required(func):
    """
    :desc: decorator method to check user's login status
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        :desc: Wrapper to check if user is logged in, if the
               stored cookies contain cookie named `acct`
               and is not expired.
        """

        is_login = False
        resp = {'success': False, 'message': 'You are not logged in!'}

        if os.path.exists(COOKIES_FILE_PATH):
            cookiejar = LWPCookieJar(filename=COOKIES_FILE_PATH)
            cookiejar.load()

            for cookie in cookiejar:
                if cookie.name == 'acct':
                    expiry_time_obj = datetime.utcfromtimestamp(cookie.expires)

                    if datetime.now() > expiry_time_obj:
                        is_login = True

            if not is_login:
                os.remove(COOKIES_FILE_PATH)
            else:
                return func(*args, **kwargs)

        return resp

    return wrapper


def get_session():
    """
    :desc: Builds session from the saved cookies, if present.
           Otherwise, a new session is created.
    :return: requests.Session object
    """

    session = Session()

    if os.path.exists(COOKIES_FILE_PATH):
        session.cookies = LWPCookieJar(filename=COOKIES_FILE_PATH)
        session.cookies.load(ignore_discard=True, ignore_expires=True)

    return session


def login_prompt():
    """
    :desc: Prompts the user to enter email and password
    :return: (email, password)
    """

    email = input('Email: ')
    password = getpass()

    return email, password


def login(email, password):
    """
    :desc: Logs a user in.
    :param: email - Email of the user - required
            password - Password of the user - required
    :return: `dict`
    """

    if email == '' or password == '':
        return {'success': False, 'message': 'Email/Password field left blank.'}

    resp = {'success': False}
    data = {'email': email, 'password': password}
    session = get_session()
    session.cookies = LWPCookieJar(filename=COOKIES_FILE_PATH)

    resp_obj = session.post(LOGIN_URL, data=data)

    if resp_obj.status_code == 200:
        if resp_obj.url == BASE_URL:
            session.cookies.save(ignore_expires=True, ignore_discard=True)
            resp['success'] = True
            resp['message'] = 'Successfully Logged In!'
        else:
            resp['message'] = 'Incorrect credentials'
    else:
        resp['message'] = 'Stackoverflow is probably down. Please try again.'

    return resp


@login_required
def logout():
    """
    :desc: Logout a user. Deletes the cookies.
    """

    session = get_session()
    logout_page_resp = session.get(LOGOUT_URL)
    resp = {'success': False}

    soup = BeautifulSoup(logout_page_resp.content, 'html.parser')
    fkey_input = soup.find('input', attrs={'name': 'fkey'})

    if fkey_input:
        data = {'fkey': fkey_input['value']}
        resp_obj = session.post(LOGOUT_URL, data=data)

        if resp_obj.url == BASE_URL:
            if os.path.exists(COOKIES_FILE_PATH):
                os.remove(COOKIES_FILE_PATH)

            resp['success'] = True
            resp['message'] = 'Successfully Logged Out!'
        else:
            resp['message'] = 'There were some problems. Please try again!'
    else:
        resp['message'] = 'There were some problems. Please try again!'

    return resp
